﻿import com.robertpenner.geom.Vector3d;

/**
* The Graphic3d class is a template for an object that contains both
* 3d coordinates and graphics. A Graphic3d object is not itself a
* MovieClip. It is a code object, which can attach and manipulate a
* MovieClip by reference.
* 
* @author Robert Penner (AS 1.0)
* © 2002 Robert Penner
* http://www.robertpenner.com
* 
* @author Mark Walters (AS 2.0)
* © 2006 DigitalFlipbook
* http://www.digitalflipbook.com
* 
* @version	1.0
* 
* @since	ActionScript 2.0; Flash Player 6
*/
class com.robertpenner.display.Graphic3d {
	
	/*
	* CLASS PROPERTIES
	*/
	private var $position:Vector3d;
	private var $screenPos:Vector3d;
	private var $timeline:MovieClip;
	private var $mc:MovieClip;
	private var $scale:Number = 100;
	
	
	/*
	* *********************************************************
	* CONSTRUCTOR
	* *********************************************************
	*
	*/
	
	/**
	* Creates an instance of the Graphic3d class.
	* 
	* @param	x			The x value of the graphic.
	* @param	y			The y value of the graphic.
	* @param	z			The z value of the graphic.
	* @param	timeline	A reference to the MovieClip that contains this particular Graphic3d object.
	* @param	mcID		The linkage name of the MovieClip symbol in the library to attach as the graphic.
	* @param	depth		An integer specifying the depth level where the graphic is placed.
	* 
	* @usage
	* <pre><code>
	* var pointA:Graphic3d = new Graphic3d(50, -30, 20, this, "dot", 1);
	* </code></pre>
	*/
	function Graphic3d (x:Number, y:Number, z:Number, timeline:MovieClip, mcID:String, depth:Number) {
		this.$position = new Vector3d (x, y, z);
		this.$timeline = timeline;
		this.$mc = this.$timeline.attachMovie (mcID, mcID + "_" + depth, depth);
		this.render ();
	}
	
	
	/*
	* *********************************************************
	* PUBLIC METHODS
	* *********************************************************
	* 
	*/
	
	/**
	* Updates the graphic's visual representation to match its current status.
	* 
	* @usage
	* <pre><code>
	* var pointA:Graphic3d = new Graphic3d(50, -30, 20, this, "dot", 1);
	* pointA.position.rotateX(5);
	* pointA.render();
	* </code></pre>
	*/
	public function render ():Void {
		var pos:Vector3d = this.$position;
		var pers:Number = pos.getPerspective();
		this.$screenPos = pos.persProjectNew (pers);
		
		this.$mc._x = this.$screenPos.x;
		this.$mc._y = -this.$screenPos.y;
		this.$mc._xscale = this.$mc._yscale = this.$scale * pers;
		this.$mc.swapDepths (100000 - pos.z * 100);
	}
	
	
	/*
	* *********************************************************
	* GETTERS/SETTERS
	* *********************************************************
	* 
	*/
	
	/**
	* The graphic's 3d location.
	* 
	* @usage
	* <pre><code>
	* var pointA:Graphic3d = new Graphic3d(50, -30, 20, this, "dot", 1);
	* pointA.position.rotateX(5);
	* </code></pre>
	*/
	public function get position ():Vector3d {
		return this.$position;
	}
	
	/**
	* The graphic's scale.
	* 
	* @usage
	* <pre><code>
	* trace (pointA.scale);
	* </code></pre>
	*/
	public function get scale ():Number {
		return this.$scale;
	}
	public function set scale (s:Number):Void {
		this.$scale = s;
	}
	
}
