﻿import com.robertpenner.utils.Degree;
import com.robertpenner.geom.Vector;

/**
* The Vector3d class is designed to represent vectors and points
* in three-dimensional space. Vectors can be added together,
* scaled, rotated, and otherwise manipulated with these methods.
* 
* @author Robert Penner (AS 1.0)
* © 2002 Robert Penner
* http://www.robertpenner.com
* 
* @author Mark Walters (AS 2.0)
* © 2006 DigitalFlipbook
* http://www.digitalflipbook.com
* 
* @version	1.0
* 
* @since	ActionScript 2.0; Flash Player 6
*/
class com.robertpenner.geom.Vector3d extends com.robertpenner.geom.Vector {
	
	/*
	* CLASS PROPERTIES
	*/
	private var $z:Number;
	
	
	/*
	* *********************************************************
	* CONSTRUCTOR
	* *********************************************************
	*
	*/
	
	/**
	* Creates an instance of the Vector3d class.
	* 
	* @param	x	The x value of the vector.
	* @param	y	The y value of the vector.
	* @param	z	The z value of the vector.
	* 
	* @usage
	* <pre><code>
	* var v:Vector3d = new Vector3d(4,0,1);
	* trace (v.x);
	* trace (v.y);
	* trace (v.z);
	* </code></pre>
	*/
	function Vector3d (x:Number, y:Number, z:Number) {
		this.initialize (x, y, z);
	}
	
	
	/*
	* *********************************************************
	* PRIVATE METHODS
	* *********************************************************
	* 
	*/
	private function initialize (x:Number, y:Number, z:Number):Void {
		super.initialize(x, y);
		this.z = z;
	}
	
	
	/*
	* *********************************************************
	* PUBLIC METHODS
	* *********************************************************
	* 
	*/
	
	/**
	* Returns a string representation of the vector object.
	* 
	* @return	Returns the string representation of the vector object.
	* 
	* @usage
	* <pre><code>
	* var position:Vector3d = new Vector3d(2,4,6.952013);
	* trace (position.toString());
	* trace (position);
	* </code></pre>
	*/
	public function toString ():String {
		var rx:Number = Math.round (this.x * 1000) / 1000;
		var ry:Number = Math.round (this.y * 1000) / 1000;
		var rz:Number = Math.round (this.z * 1000) / 1000;
		return "[" + rx + ", " + ry + ", " + rz + "]";
	}
	
	/**
	* Reinitializes the vector object.
	* 
	* @param	x	The x value of the vector.
	* @param	y	The y value of the vector.
	* @param	z	The z value of the vector.
	* 
	* @usage
	* <pre><code>
	* var acceleration:Vector3d = new Vector3d(0,0,0);
	* acceleration.reset(10,20,30);
	* trace (acceleration);
	* </code></pre>
	*/
	public function reset (x:Number, y:Number, z:Number):Void {
		this.initialize (x, y, z);
	}
	
	/**
	* Returns a new vector object containing the x, y and z values of the current vector.
	* 
	* @return	Returns a copy of the current vector object.
	* 
	* @usage
	* <pre><code>
	* var force:Vector3d = new Vector3d(2,4,6);
	* var forceCopy:Vector3d = force.getClone();
	* trace (forceCopy);
	* trace (forceCopy == force);
	* </code></pre>
	*/
	public function getClone ():Vector3d {
		return new Vector3d (this.x, this.y, this.z);
	}
	
	/**
	* Tests the equality of two vector objects.
	* 
	* @param	v	The vector object to compare with the current vector.
	* @return	Returns true if the values of the two x properties, the two y properties and the two z properties are the same.
	* 
	* @usage
	* <pre><code>
	* var force:Vector3d = new Vector3d(2,4,6);
	* var forceCopy:Vector3d = force.getClone();
	* trace (forceCopy.equals(force));
	* </code></pre>
	*/
	public function equals (v:Vector3d):Object {
		return (super.equals(v) && this.z == v.z);
	}
	
	/**
	* Adds a vector object to the current vector.
	* 
	* @param	v	The vector object to add to the current vector.
	* 
	* @usage
	* <pre><code>
	* var position:Vector3d = new Vector3d(1,1,1);
	* var velocity:Vector3d = new Vector3d(2,1,3);
	* position.plus(velocity);
	* trace (position);
	* </code></pre>
	*/
	public function plus (v:Vector3d):Void {
		super.plus(v);
		this.z += v.z;
	}
	
	/**
	* Performs addition on two vector objects and returns the result of the addition as a new vector.
	* 
	* @param	v	The vector object to add to the current vector.
	* @return	Returns the result of the addition as a new vector.
	* 
	* @usage
	* <pre><code>
	* var position:Vector3d = new Vector3d(1,1,1);
	* var velocity:Vector3d = new Vector3d(2,1,3);
	* var newPosition:Vector3d = position.plusNew(velocity);
	* trace (newPosition);
	* </code></pre>
	*/
	public function plusNew (v:Vector3d):Vector3d {
		return new Vector3d (this.x + v.x, this.y + v.y, this.z + v.z);
	}
	
	/**
	* Subtracts a vector object from the current vector.
	* 
	* @param	v	The vector object to subtract from the current vector.
	* 
	* @usage
	* <pre><code>
	* var farPoint:Vector3d = new Vector3d(10,-6,5);
	* var nearPoint:Vector3d = new Vector3d(6,-2,0);
	* farPoint.minus(nearPoint);
	* trace (farPoint);
	* </code></pre>
	*/
	public function minus (v:Vector3d):Void {
		super.minus(v);
		this.z -= v.z;
	}
	
	/**
	* Performs subtraction on two vector objects and returns the result of the subtraction as a new vector.
	* 
	* @param	v	The vector object to subtract from the current vector.
	* @return	Returns the result of the subtraction as a new vector.
	* 
	* @usage
	* <pre><code>
	* var farPoint:Vector3d = new Vector3d(10,-6,5);
	* var nearPoint:Vector3d = new Vector3d(6,-2,0);
	* var displacement:Vector3d = farPoint.minusNew(nearPoint);
	* trace (displacement);
	* </code></pre>
	*/
	public function minusNew (v:Vector3d):Vector3d {
		return new Vector3d (this.x - v.x, this.y - v.y, this.z - v.z);
	}
	
	/**
	* Reverses the direction of the current vector.
	* 
	* @usage
	* <pre><code>
	* var direction:Vector3d = new Vector3d(3,-6,-9);
	* direction.negate();
	* trace (direction);
	* </code></pre>
	*/
	public function negate ():Void {
		super.negate();
		this.z = - z;
	}
	
	/**
	* Reverses the direction of the current vector and returns the result as a new vector.
	* 
	* @return	Returns the reversed vector object as a new vector.
	* 
	* @usage
	* <pre><code>
	* var towards:Vector3d = new Vector3d(0,0,-20);
	* var away:Vector3d = towards.negateNew();
	* trace (away);
	* </code></pre>
	*/
	public function negateNew ():Vector3d {
		return new Vector3d (-this.x, -this.y, -this.z);
	}
	
	/**
	* Scales the length of the current vector object by a scale factor.
	* 
	* @param	s	The scale factor to multiply the current vector by.
	* 
	* @usage
	* <pre><code>
	* var tensionForce:Vector3d = new Vector3d(3,-1,0);
	* tensionForce.scale(2);
	* trace (tensionForce);
	* </code></pre>
	*/
	public function scale (s:Number):Void {
		super.scale(s);
		this.z *= s;
	}
	
	/**
	* Scales the length of the current vector object by a scale factor and returns the result of the scale as a new vector.
	* 
	* @param	s	The scale factor to multiply the current vector by.
	* @return	Returns the scaled vector object as a new vector.
	* 
	* @usage
	* <pre><code>
	* var tensionForce:Vector3d = new Vector3d(3,-1,0);
	* var strongerForce:Vector3d = tensionForce.scaleNew(2);
	* trace (strongerForce);
	* </code></pre>
	*/
	public function scaleNew (s:Number):Vector3d {
		return new Vector3d (this.x * s, this.y * s, this.z * s);
	}
	
	/**
	* Gets the size of the current vector.
	* 
	* @return	Returns the size of the current vector.
	* 
	* @usage
	* <pre><code>
	* var velocity:Vector3d = new Vector3d(3,4,0);
	* var speed:Number = velocity.getLength();
	* trace (speed);
	* </code></pre>
	*/
	public function getLength ():Number {
		return Math.sqrt (this.x*this.x + this.y*this.y + this.z*this.z);
	}
	
	/**
	* Sets the size of the current vector.
	* 
	* @param	len		The length that the current vector will be set.
	* 
	* @usage
	* <pre><code>
	* var velocity:Vector3d = new Vector3d(3,4,0);
	* var newSpeed:Number = 10;
	* velocity.setLength(newSpeed);
	* trace (velocity);
	* trace (velocity.getLength());
	* </code></pre>
	*/
	public function setLength (len:Number):Void {
		super.setLength(len);
	}
	
	public function getAngle ():Number {
		if (this.z == 0) {
			return super.getAngle();
		} else {
			trace ("ERROR: No getAngle() method in Vector3d class");
			return undefined;
		}
	}
	
	public function setAngle (ang:Number):Void {
		if (this.z == 0) {
			super.setAngle(ang);
		} else {
			trace ("ERROR: No setAngle() method in Vector3d class");
		}
	}
	
	public function rotate (ang:Number):Void {
		if (this.z == 0) {
			super.rotate(ang);
		} else {
			trace ("ERROR: No rotate() method in Vector3d class");
		}
	}
	
	public function rotateNew (ang:Number):Vector3d {
		if (this.z == 0) {
			var rv:Vector = super.rotateNew(ang);
			return new Vector3d (rv.x, rv.y, this.z);
		} else {
			trace ("ERROR: No rotateNew() method in Vector3d class");
		}
	}
	
	/**
	* Performs multiplication on two vector objects and returns the dot product.
	* 
	* @param	v	The vector object to multiply the current vector by.
	* @return	Returns the dot product.
	* 
	* @usage
	* <pre><code>
	* var v:Vector3d = new Vector3d(2,0,1);
	* var w:Vector3d = new Vector3d(3,5,4);
	* trace (v.dot(w));
	* </code></pre>
	*/
	public function dot (v:Vector3d):Number {
		return super.dot(v) + this.z * v.z;
	}
	
	/**
	* Performs multiplication on two vector objects and returns a third vector object that is perpendicular to the first two vectors.
	* 
	* @param	v	The vector object to multiply the current vector by.
	* @return	Returns a normal vector to the other two vector objects.
	* 
	* @usage
	* <pre><code>
	* var a:Vector3d = new Vector3d(2,0,0);
	* var b:Vector3d = new Vector3d(0,2,0);
	* var c:Vector3d = a.cross(b);
	* trace (c);
	* var d:Vector3d = b.cross(a);
	* trace (d);
	* </code></pre>
	*/
	public function cross (v:Vector3d):Vector3d {
		var cx:Number = this.y * v.z - this.z * v.y;
		var cy:Number = this.z * v.x - this.x * v.z;
		var cz:Number = this.x * v.y - this.y * v.x;
		return new Vector3d (cx, cy, cz);
	}
	
	public function getNormal ():Vector3d {
		if (this.z == 0) {
			var normal:Vector = super.getNormal();
			return new Vector3d (normal.x, normal.y, this.z);
		} else {
			trace ("ERROR: No getNormal() method in Vector3d class");
			return undefined;
		}
	}
	
	public function isPerpTo (v:Vector3d):Boolean {
		if (this.z == 0 && v.z == 0) {
			return super.isPerpTo(v);
		} else {
			trace ("ERROR: No isPerpTo() method in Vector3d class");
			return undefined;
		}
	}
	
	public function isNormalTo (v:Vector3d):Boolean {
		if (this.z == 0 && v.z == 0) {
			return this.isPerpTo(v);
		} else {
			trace ("ERROR: No isNormalTo() method in Vector3d class");
			return undefined;
		}
	}
	
	/**
	* Returns the angle between the current vector and another vector.
	* 
	* @param	v	The vector object to check the angle between.
	* @return	Returns the angle between the current vector and another vector.
	* 
	* @usage
	* <pre><code>
	* var pullForce:Vector3d = new Vector3d(4,0,0);
	* var frictionForce:Vector3d = new Vector3d(-1,0,1);
	* var theta:Number = pullForce.angleBetween(frictionForce);
	* trace (theta);
	* </code></pre>
	*/
	public function angleBetween (v:Vector):Number {
		return super.angleBetween(v);
	}
	
	/**
	* Calculates and returns the perspective ratio needed to scale an object correctly.
	* 
	* @param	viewDist	The viewing distance of the projection.<br />
	* 						Large values produce an image in which there is little size difference between close and distant objects.<br />
	* 						Small values produce more of a "fisheye" effect, distorting the image like a peephole.<br />
	* 						If no value is specified, a default value of 300 is chosen, which gives a natural-looking perspective.
	* @return	Returns the perspective ratio needed to scale an object correctly.
	* 
	* @usage
	* <pre><code>
	* var pointA:Vector3d = new Vector3d(50,20,40);
	* var pers:Number = pointA.getPerspective();
	* trace (pers);
	* </code></pre>
	*/
	public function getPerspective (viewDist:Number):Number {
		viewDist = viewDist ? viewDist : 300;
		return viewDist / (this.z + viewDist);
	}
	
	/**
	* Performs a perspective projection on a 3d point. It converts (x, y, z) coordinates to a 2d location (x, y) on the screen.
	* 
	* @param	p	The perspective ratio. If no value is specified, it is calculated automatically by calling the getPerspective() method.
	* 
	* @usage
	* <pre><code>
	* var pointA:Vector3d = new Vector3d(50,20,40);
	* pointA.persProject();
	* trace (pointA);
	* </code></pre>
	*/
	public function persProject (p:Number):Void {
		p = p ? p : this.getPerspective();
		this.x *= p;
		this.y *= p;
		this.z = 0;
	}
	
	/**
	* Performs a perspective projection on a 3d point and returns the result of the projection as a new vector. It converts (x, y, z) coordinates to a 2d location (x, y) on the screen.
	* 
	* @param	p	The perspective ratio. If no value is specified, it is calculated automatically by calling the getPerspective() method.
	* @return	Returns the result of the projection as a new vector.
	* 
	* @usage
	* <pre><code>
	* var pointA:Vector3d = new Vector3d(50,20,40);
	* pointA.persProject();
	* trace (pointA);
	* </code></pre>
	*/
	public function persProjectNew (p:Number):Vector3d {
		p = p ? p : this.getPerspective();
		return new Vector3d (p*this.x, p*this.y, 0);
	}
	
	/**
	* Rotates the current vector object around the x-axis by a certain amount of degrees.
	* 
	* @param	ang		The amount of degrees that the current vector object will be rotated by.
	* 
	* @usage
	* <pre><code>
	* var p:Vector3d = new Vector3d(1,4,7);
	* p.rotateX(180);
	* trace (p);
	* </code></pre>
	*/
	public function rotateX (ang:Number):Void {
		var ca:Number = Degree.cosD (ang);
		var sa:Number = Degree.sinD (ang);
		this.rotateXTrig(ca, sa);
	}
	
	/**
	* Rotates the current vector object around the x-axis by the cosine and sine of an angle.
	* 
	* @param	ca		The cosine of the angle to rotate the current vector object by.
	* @param	sa		The sine of the angle to rotate the current vector object by.
	* 
	* @usage
	* <pre><code>
	* var p:Vector3d = new Vector3d(1,4,7);
	* var cosAngle:Number = Degree.cosD(180);
	* var sinAngle:Number = Degree.sinD(180);
	* p.rotateXTrig(cosAngle, sinAngle);
	* trace (p);
	* </code></pre>
	*/
	public function rotateXTrig (ca:Number, sa:Number):Void {
		var ry:Number = this.y * ca - this.z * sa;
		var rz:Number = this.y * sa + this.z * ca;
		this.y = ry;
		this.z = rz;
	}
	
	/**
	* Rotates the current vector object around the y-axis by a certain amount of degrees.
	* 
	* @param	ang		The amount of degrees that the current vector object will be rotated by.
	* 
	* @usage
	* <pre><code>
	* var p:Vector3d = new Vector3d(1,4,7);
	* p.rotateY(180);
	* trace (p);
	* </code></pre>
	*/
	public function rotateY (ang:Number):Void {
		var ca:Number = Degree.cosD (ang);
		var sa:Number = Degree.sinD (ang);
		this.rotateYTrig(ca, sa);
	}
	
	/**
	* Rotates the current vector object around the y-axis by the cosine and sine of an angle.
	* 
	* @param	ca		The cosine of the angle to rotate the current vector object by.
	* @param	sa		The sine of the angle to rotate the current vector object by.
	* 
	* @usage
	* <pre><code>
	* var p:Vector3d = new Vector3d(3,-8,5);
	* var cosAngle:Number = Degree.cosD(90);
	* var sinAngle:Number = Degree.sinD(90);
	* p.rotateYTrig(cosAngle, sinAngle);
	* trace (p);
	* </code></pre>
	*/
	public function rotateYTrig (ca:Number, sa:Number):Void {
		var rx:Number = this.x * ca + this.z * sa;
		var rz:Number = this.x * -sa + this.z * ca;
		this.x = rx;
		this.z = rz;
	}
	
	/**
	* Rotates the current vector object around the z-axis by a certain amount of degrees.
	* 
	* @param	ang		The amount of degrees that the current vector object will be rotated by.
	* 
	* @usage
	* <pre><code>
	* var p:Vector3d = new Vector3d(1,4,7);
	* p.rotateZ(180);
	* trace (p);
	* </code></pre>
	*/
	public function rotateZ (ang:Number):Void {
		var ca:Number = Degree.cosD (ang);
		var sa:Number = Degree.sinD (ang);
		this.rotateZTrig(ca, sa);
	}
	
	/**
	* Rotates the current vector object around the z-axis by the cosine and sine of an angle.
	* 
	* @param	ca		The cosine of the angle to rotate the current vector object by.
	* @param	sa		The sine of the angle to rotate the current vector object by.
	* 
	* @usage
	* <pre><code>
	* var p:Vector3d = new Vector3d(6,1,4);
	* var cosAngle:Number = Degree.cosD(45);
	* var sinAngle:Number = Degree.sinD(45);
	* p.rotateZTrig(cosAngle, sinAngle);
	* trace (p);
	* </code></pre>
	*/
	public function rotateZTrig (ca:Number, sa:Number):Void {
		var rx:Number = this.x * ca - this.y * sa;
		var ry:Number = this.x * sa + this.y * ca;
		this.x = rx;
		this.y = ry;
	}
	
	/**
	* Rotates the current vector object around the x and y axes by a certain amount of degrees.
	* 
	* @param	a		The amount of degrees that the current vector object will be rotated around the x-axis by.
	* @param	b		The amount of degrees that the current vector object will be rotated around the y-axis by.
	* 
	* @usage
	* <pre><code>
	* var p:Vector3d = new Vector3d(8,0,0);
	* p.rotateXY(45,45);
	* trace (p);
	* </code></pre>
	*/
	public function rotateXY (a:Number, b:Number):Void {
		var ca:Number = Degree.cosD (a);
		var sa:Number = Degree.sinD (a);
		var cb:Number = Degree.cosD (b);
		var sb:Number = Degree.sinD (b);
		this.rotateXYTrig(ca, sa, cb, sb);
	}
	
	/**
	* Rotates the current vector object around the x and y axes by the cosine and sine of an angle.
	* 
	* @param	ca		The cosine of the angle to rotate the current vector object around the x-axis by.
	* @param	sa		The sine of the angle to rotate the current vector object around the x-axis by.
	* @param	cb		The cosine of the angle to rotate the current vector object around the y-axis by.
	* @param	sb		The sine of the angle to rotate the current vector object around the y-axis by.
	* 
	* @usage
	* <pre><code>
	* var p:Vector3d = new Vector3d(6,1,4);
	* var cosAngleA:Number = Degree.cosD(45);
	* var sinAngleA:Number = Degree.sinD(45);
	* var cosAngleB:Number = Degree.cosD(90);
	* var sinAngleB:Number = Degree.sinD(90);
	* p.rotateXYTrig(cosAngleA, sinAngleA, cosAngleB, sinAngleB);
	* trace (p);
	* </code></pre>
	*/
	public function rotateXYTrig (ca:Number, sa:Number, cb:Number, sb:Number):Void {
		// x-axis rotation
		var rz:Number = this.y * sa + this.z * ca;
		this.y = this.y * ca - this.z * sa;
		// y-axis rotation
		this.z = this.x * -sb + rz * cb;
		this.x = this.x * cb + rz * sb;
	}
	
	/**
	* Rotates the current vector object around the x, y and z axes by a certain amount of degrees.
	* 
	* @param	a		The amount of degrees that the current vector object will be rotated around the x-axis by.
	* @param	b		The amount of degrees that the current vector object will be rotated around the y-axis by.
	* @param	c		The amount of degrees that the current vector object will be rotated around the z-axis by.
	* 
	* @usage
	* <pre><code>
	* var p:Vector3d = new Vector3d(8,0,0);
	* p.rotateXYZ(45,45,45);
	* trace (p);
	* </code></pre>
	*/
	public function rotateXYZ (a:Number, b:Number, c:Number):Void {
		var ca:Number = Degree.cosD (a);
		var sa:Number = Degree.sinD (a);
		var cb:Number = Degree.cosD (b);
		var sb:Number = Degree.sinD (b);
		var cc:Number = Degree.cosD (c);
		var sc:Number = Degree.sinD (c);
		this.rotateXYZTrig(ca, sa, cb, sb, cc, sc);
	}
	
	/**
	* Rotates the current vector object around the x, y and z axes by the cosine and sine of an angle.
	* 
	* @param	ca		The cosine of the angle to rotate the current vector object around the x-axis by.
	* @param	sa		The sine of the angle to rotate the current vector object around the x-axis by.
	* @param	cb		The cosine of the angle to rotate the current vector object around the y-axis by.
	* @param	sb		The sine of the angle to rotate the current vector object around the y-axis by.
	* @param	cc		The cosine of the angle to rotate the current vector object around the z-axis by.
	* @param	sc		The sine of the angle to rotate the current vector object around the z-axis by.
	* 
	* @usage
	* <pre><code>
	* var p:Vector3d = new Vector3d(6,1,4);
	* var cosAngleA:Number = Degree.cosD(45);
	* var sinAngleA:Number = Degree.sinD(45);
	* var cosAngleB:Number = Degree.cosD(90);
	* var sinAngleB:Number = Degree.sinD(90);
	* var cosAngleC:Number = Degree.cosD(135);
	* var sinAngleC:Number = Degree.sinD(135);
	* p.rotateXYZTrig(cosAngleA, sinAngleA, cosAngleB, sinAngleB, cosAngleC, sinAngleC);
	* trace (p);
	* </code></pre>
	*/
	public function rotateXYZTrig (ca:Number, sa:Number, cb:Number, sb:Number, cc:Number, sc:Number):Void {
		// x-axis rotation
		var ry:Number = this.y * ca - this.z * sa;
		var rz:Number = this.y * sa + this.z * ca;
		// y-axis rotation
		var rx:Number = this.x * cb + rz * sb;
		this.z = this.x * -sb + rz * cb;
		// z-axis rotation
		this.x = rx * cc - ry * sc;
		this.y = rx * sc + ry * cc;
	}
	
	
	/*
	* *********************************************************
	* GETTERS/SETTERS
	* *********************************************************
	* 
	*/
	
	/**
	* The x property of the vector object.
	* 
	* @usage
	* <pre><code>
	* v.x = -4;
	* </code></pre>
	*/
	public function get x ():Number {
		return this.$x;
	}
	public function set x (x:Number):Void {
		this.$x = x;
	}
	
	/**
	* The y property of the vector object.
	* 
	* @usage
	* <pre><code>
	* v.y = 0;
	* </code></pre>
	*/
	public function get y ():Number {
		return this.$y;
	}
	public function set y (y:Number):Void {
		this.$y = y;
	}
	
	/**
	* The z property of the vector object.
	* 
	* @usage
	* <pre><code>
	* v.z = -1;
	* </code></pre>
	*/
	public function get z ():Number {
		return this.$z;
	}
	public function set z (z:Number):Void {
		this.$z = z;
	}
	
	/**
	* The length of the vector object.
	* 
	* @usage
	* <pre><code>
	* var velocity:Vector3d = new Vector3d(3,4,0);
	* var newSpeed:Number = 10;
	* velocity.length = newSpeed;
	* trace (velocity);
	* trace (velocity.length);
	* </code></pre>
	*/
	public function get length ():Number {
		return this.getLength();
	}
	public function set length (len:Number):Void {
		this.setLength(len);
	}
}
